//+------------------------------------------------------------------+
//|                                              ClusterX_Volume.mq4 |
//+------------------------------------------------------------------+
#property copyright "Copyright 2013, TheXpert"
#property link      "http://trend-lab.blogspot.com/"

#define VERSION_MAJOR 0
#define VERSION_MINOR 3
#define MY_NAME "ClusterX_Volume"

#import "ntdll.dll"
   int RtlGetLastWin32Error();
   int RtlSetLastWin32Error(int dwErrCode);

#import "kernel32.dll"
   int   DeleteFileA (string lpFileName);
   int   CloseHandle(int hObject);
   int   CreateFileA(     string lpFileName,         int dwDesiredAccess, 
                           int dwShareMode,           int lpSecurityAttributes,
                           int dwCreationDisposition, int dwFlagsAndAttributes, 
                           int hTemplateFile
                     );

   int   SetFilePointer(int hFile, int lDistanceToMove, int& lpDistanceToMoveHigh[], int dwMoveMethod);
   int   ReadFile(int hFile, double& lpBuffer[], int nNumberOfBytesToRead, int& lpNumberOfBytesRead[], int lpOverlapped);
   int   WriteFile(int  hFile, string lpBuffer, int nNumberOfBytesToWrite, int& lpNumberOfBytesWritten[], int  lpOverlapped);
#import

#define F_GENERIC_WRITE                0x40000000 //     .
#define F_GENERIC_READ                 0x80000000 //     .

// ----  "dwShareMode".
#define F_FILE_SHARE_NO                0x00000000 //    .
#define F_FILE_SHARE_READ              0x00000001 //          .
#define F_FILE_SHARE_WRITE             0x00000002 //          .

// ----  "dwCreationDisposition".
#define F_CREATE_ALWAYS                2          //   , .
#define F_OPEN_EXISTING                3          //    ,   .
#define F_OPEN_ALWAYS                  4          //  , .

//====    "SetFilePointer()".
#define F_FILE_BEGIN                   0          //   .

#define INVALID_SET_FILE_POINTER -1
#define INVALID_FILE_ATTRIBUTES -1
#define INVALID_HANDLE_VALUE -1
#define INVALID_FILE_SIZE 0xFFFFFFFF

#define ERROR_FILE_NOT_FOUND 2
#define ERROR_PATH_NOT_FOUND 3


#property indicator_separate_window

#property indicator_buffers 1
#property indicator_color1 Gray

#property indicator_width1 2

#property indicator_minimum 0

extern string Ticker = "AUTO";

string VolumePath;
string DataFilePrefix;

datetime LastAvailable;

int CurrentHandle;
int LockHandle;
string CurrentName;
string LockName;

double Values[];

void CheckReceiverAvailable()
{
   RtlSetLastWin32Error(0);
   bool succeeded = DeleteFileA(TerminalPath() + "\\ClusterDelta\\VolumeDataReceiver");
   int err = RtlGetLastWin32Error();
   
   if (succeeded || err == ERROR_FILE_NOT_FOUND || err == ERROR_PATH_NOT_FOUND)
   { // receiver is not running
      Alert("Receiver for volume is not running. Run volume receiver to receive data online");
   }
}

bool FirstRun;
bool NextTick = false;
datetime NextTime;

int init()
{
   FirstRun = true;
   NextTick = false;
   NextTime = 0;
   
   SetIndexBuffer(0, Values);
   SetIndexStyle(0, DRAW_HISTOGRAM);
   
   VolumePath = TerminalPath() + "\\ClusterDelta\\Volume\\Data\\";
   DataFilePrefix = VolumePath + Symbol() + "_" + Ticker + "_" + EncodePeriod(Period()) + "_";
   
   LastAvailable = 0;
   
   CurrentHandle = INVALID_HANDLE_VALUE;
   LockHandle = INVALID_HANDLE_VALUE;
   
   LockName = TerminalPath() + "\\ClusterDelta\\Volume\\" + EncodePeriod(Period()) + Ticker + "_" + Symbol();

   TryToCaptureFile(LockName);
   
   return(0);
}

int start()
{
   if (NextTick && NextTime < TimeLocal())
   {
      NextTick = false;
      
      CheckReceiverAvailable();
   }
   
   if (FirstRun)
   {
      FirstRun = false;
      ArrayInitialize(Values, EMPTY_VALUE);
      
      NextTick = true;
      NextTime = TimeLocal() + 20;
   }

   int startTime = GetTickCount();
   
   TryToCaptureFile(LockName);

   bool triedThisTick = false;
   
   int toCount = Bars - IndicatorCounted();
   
   int last = iBarShift(Symbol(), Period(), LastAvailable) + 1;
   
   if (last > toCount) toCount = last;
   
   for (int i = toCount - 1; i >= 0; i--)
   {
      string name = TimeYear(Time[i]) + "_" + TimeMonth(Time[i]);
      string fullPath = DataFilePrefix + name;

      if (fullPath == CurrentName)
      {
         // if names are equal maybe we have needed handle
         if (CurrentHandle == INVALID_HANDLE_VALUE)
         {
            // if no handle is open for now
            if (triedThisTick)
            {
               // not for this time
               continue;
            }
            else
            {
               RtlSetLastWin32Error(0);
               int hFile = CreateFileA(CurrentName, F_GENERIC_READ , F_FILE_SHARE_READ | F_FILE_SHARE_WRITE, 0, F_OPEN_EXISTING, 0, 0);
   
               if (INVALID_HANDLE_VALUE == hFile)
               {
                  int err = RtlGetLastWin32Error();
                  triedThisTick = true;

                  if (err != ERROR_FILE_NOT_FOUND && err != ERROR_PATH_NOT_FOUND)
                  {
                     Print("Unexpected error opening file, err = ", err);
                  }
                  continue;
               }
               
               CurrentHandle = hFile;
            }
         }
      }
      else
      {
         if (CurrentHandle != INVALID_HANDLE_VALUE)
         {
            CloseHandle(CurrentHandle);
            CurrentHandle = INVALID_HANDLE_VALUE;
         }
         
         CurrentName = fullPath;
         
         RtlSetLastWin32Error(0);
         hFile = CreateFileA(CurrentName, F_GENERIC_READ , F_FILE_SHARE_READ | F_FILE_SHARE_WRITE, 0, F_OPEN_EXISTING, 0, 0);

         if (INVALID_HANDLE_VALUE == hFile)
         {
            err = RtlGetLastWin32Error();
            triedThisTick = true;

            if (err != ERROR_FILE_NOT_FOUND && err != ERROR_PATH_NOT_FOUND)
            {
               Print("Unexpected error opening file, err = ", err);
            }
            continue;
         }
         CurrentHandle = hFile;
      }
      
      // if we are here, we assume that CurrentHandle is available and correct
      // counting offset
      
      datetime t = Time[i];
      int offset = (((TimeDay(t) - 1)*24*60 + TimeHour(t)*60 + TimeMinute(t))/Period())*8;
      int hi[] = {0};
      
      RtlSetLastWin32Error(0);
      int newPos = SetFilePointer(CurrentHandle, offset, hi, 0);
      if (newPos == INVALID_SET_FILE_POINTER)
      {
         err = RtlGetLastWin32Error();
         Print("Should not happen! Set pos when reading Delta failed, err = ", err);
         
         if (err == 6)
         {
            CloseHandle(CurrentHandle);
            CurrentHandle = INVALID_HANDLE_VALUE;
         }
         break;
      }
      
      double data[] = {0};
      int read[] = {0};
      
      RtlSetLastWin32Error(0);
      if (0 != ReadFile(CurrentHandle, data, 8, read, 0))
      {
         // we have read data successfully
         if (data[0] != EMPTY_VALUE)
         {
            LastAvailable = Time[i];
            Values[i] = data[0];
         }
      }
      else
      {
         Print("ReadFile failed");
      }
   }
   
   return(0);
}

void deinit()
{
   if (INVALID_HANDLE_VALUE != CurrentHandle)
   {
      CloseHandle(CurrentHandle);
      CurrentHandle = INVALID_HANDLE_VALUE;
   }

   if (INVALID_HANDLE_VALUE != LockHandle)
   {
      CloseHandle(LockHandle);
      LockHandle = INVALID_HANDLE_VALUE;
   }
}

string EncodePeriod(int tf)
{
   switch (tf)
   {
      case PERIOD_M1    : return ("M01");
      case PERIOD_M5    : return ("M05");
      case PERIOD_M15   : return ("M15");
      case PERIOD_M30   : return ("M30");
      case PERIOD_H1    : return ("H01");
      case PERIOD_H4    : return ("H04");
      case PERIOD_D1    : return ("D01");
      case PERIOD_W1    : return ("W01");
      case PERIOD_MN1   : return ("MN1");
   }
   return ("n/a");
}

void TryToCaptureFile(string fileName)
{
   if (LockHandle != INVALID_HANDLE_VALUE) return;
   
   int hFile = CreateFileA(fileName, F_GENERIC_WRITE | F_GENERIC_READ, F_FILE_SHARE_NO, 0, F_CREATE_ALWAYS, 0, 0);
   
   bool succeeded = false;
   if (INVALID_HANDLE_VALUE != hFile)
   {
      int written[] = {0};
      succeeded = (WriteFile(hFile, Ticker, StringLen(Ticker), written, 0) != 0);

      CloseHandle(hFile);
   }
   
   if (succeeded)
   {
      hFile = CreateFileA(fileName, F_GENERIC_READ, F_FILE_SHARE_READ, 0, F_OPEN_EXISTING, 0, 0);

      if (INVALID_HANDLE_VALUE != hFile)
      {
         LockHandle = hFile;
      }
   }
}